import argparse
import logging
import os
import sqlite3
from array import array
from dataclasses import dataclass, field
from mmap import PAGESIZE
from pathlib import Path
from typing import Any, List

from pyxcp.recorder import XcpLogFileDecoder
from pyxcp.recorder.converter import MAP_TO_ARRAY


MAP_TO_SQL = {
    "U8": "INTEGER",
    "I8": "INTEGER",
    "U16": "INTEGER",
    "I16": "INTEGER",
    "U32": "INTEGER",
    "I32": "INTEGER",
    "U64": "INTEGER",
    "I64": "INTEGER",
    "F32": "FLOAT",
    "F64": "FLOAT",
    "F16": "FLOAT",
    "BF16": "FLOAT",
}

logger = logging.getLogger("PyXCP")

parser = argparse.ArgumentParser(description="Use .xmraw files in an Apache Arrow application.")
parser.add_argument("xmraw_file", help=".xmraw file")
args = parser.parse_args()


@dataclass
class Storage:
    name: str
    arrow_type: Any
    arr: array


@dataclass
class StorageContainer:
    name: str
    arr: List[Storage] = field(default_factory=[])
    ts0: List[int] = field(default_factory=lambda: array("Q"))
    ts1: List[int] = field(default_factory=lambda: array("Q"))


class Decoder(XcpLogFileDecoder):

    def __init__(self, recording_file_name: str):
        super().__init__(recording_file_name)
        self.sq3_file_name = Path(recording_file_name).with_suffix(".sq3")
        try:
            os.unlink(self.sq3_file_name)
        except Exception as e:
            print(e)

    def initialize(self) -> None:
        self.create_database(self.sq3_file_name)
        self.arrow_tables = []
        self.insert_stmt = {}
        for dl in self.daq_lists:
            result = []
            for name, type_str in dl.headers:
                array_txpe = MAP_TO_ARRAY[type_str]
                sql_type = MAP_TO_SQL[type_str]
                sd = Storage(name, sql_type, array(array_txpe))
                result.append(sd)
            sc = StorageContainer(dl.name, result)
            print(f"Creating table {sc.name!r}.")
            self.create_table(sc)
            self.insert_stmt[sc.name] = (
                f"""INSERT INTO {sc.name}({', '.join(['ts0', 'ts1'] + [r.name for r in sc.arr])}) VALUES({', '.join(["?" for _ in range(len(sc.arr) + 2)])})"""
            )
            self.arrow_tables.append(sc)
        print("\nInserting data...")

    def create_database(self, db_name: str) -> None:
        self.conn = sqlite3.Connection(db_name)
        self.cursor = self.conn.cursor()
        self.execute("PRAGMA FOREIGN_KEYS=ON")
        self.execute(f"PRAGMA PAGE_SIZE={PAGESIZE}")
        self.execute("PRAGMA SYNCHRONOUS=OFF")
        self.execute("PRAGMA LOCKING_MODE=EXCLUSIVE")
        self.execute("PRAGMA TEMP_STORE=MEMORY")

        timestamp_info = self.parameters.timestamp_info
        self.execute(
            "CREATE TABLE timestamp_info(timestamp_ns INTEGER, utc_offset INTEGER, dst_offset INTEGER, timezone VARCHAR(255))"
        )
        self.execute("CREATE TABLE table_names(name VARCHAR(255))")
        self.execute(
            "INSERT INTO timestamp_info VALUES(?, ?, ?, ?)",
            [timestamp_info.timestamp_ns, timestamp_info.utc_offset, timestamp_info.dst_offset, timestamp_info.timezone],
        )

    def create_table(self, sc: StorageContainer) -> None:
        columns = ["ts0 INTEGER", "ts1 INTEGER"]
        for elem in sc.arr:
            columns.append(f"{elem.name} {elem.arrow_type}")
        ddl = f"CREATE TABLE {sc.name}({', '.join(columns)})"
        self.execute(ddl)
        self.execute("INSERT INTO table_names VALUES(?)", [sc.name])

    def execute(self, *args: List[str]) -> None:
        try:
            self.cursor.execute(*args)
        except Exception as e:
            print(e)

    def finalize(self) -> None:
        self.conn.commit()
        self.conn.close()
        print("Done.")

    def on_daq_list(self, daq_list_num: int, timestamp0: int, timestamp1: int, measurements: list) -> None:
        sc = self.arrow_tables[daq_list_num]
        insert_stmt = self.insert_stmt[sc.name]
        data = [timestamp0, timestamp1, *measurements]
        self.execute(insert_stmt, data)


decoder = Decoder(args.xmraw_file)
decoder.run()



#!/usr/bin/env python
from .base import FrameAcquisitionPolicy  # noqa: F401
from .base import FrameRecorderPolicy  # noqa: F401
from .base import LegacyFrameAcquisitionPolicy  # noqa: F401
from .base import NoOpPolicy  # noqa: F401
from .base import StdoutPolicy  # noqa: F401
from .can import Can  # noqa: F401
from .eth import Eth  # noqa: F401
from .sxi import SxI  # noqa: F401
from .usb_transport import Usb  # noqa: F401



import argparse
import logging
import os
import csv
from array import array
from dataclasses import dataclass, field
from mmap import PAGESIZE
from pathlib import Path
from typing import Any, List

from pyxcp.recorder import XcpLogFileDecoder
from pyxcp.recorder.converter import MAP_TO_ARRAY


MAP_TO_SQL = {
    "U8": "INTEGER",
    "I8": "INTEGER",
    "U16": "INTEGER",
    "I16": "INTEGER",
    "U32": "INTEGER",
    "I32": "INTEGER",
    "U64": "INTEGER",
    "I64": "INTEGER",
    "F32": "FLOAT",
    "F64": "FLOAT",
    "F16": "FLOAT",
    "BF16": "FLOAT",
}

logger = logging.getLogger("PyXCP")

parser = argparse.ArgumentParser(description="Use .xmraw files in an Apache Arrow application.")
parser.add_argument("xmraw_file", help=".xmraw file")
args = parser.parse_args()


@dataclass
class Storage:
    name: str
    arrow_type: Any
    arr: array


@dataclass
class StorageContainer:
    name: str
    arr: List[Storage] = field(default_factory=[])
    ts0: List[int] = field(default_factory=lambda: array("Q"))
    ts1: List[int] = field(default_factory=lambda: array("Q"))


class Decoder(XcpLogFileDecoder):

    def __init__(self, recording_file_name: str):
        super().__init__(recording_file_name)

    def initialize(self) -> None:
        self.arrow_tables = []
        self.csv_writers = []
        for dl in self.daq_lists:
            result = []
            for name, type_str in dl.headers:
                array_txpe = MAP_TO_ARRAY[type_str]
                sql_type = MAP_TO_SQL[type_str]
                sd = Storage(name, sql_type, array(array_txpe))
                result.append(sd)
            sc = StorageContainer(dl.name, result)
            writer = csv.writer(open(f"{sc.name}.csv", "w", newline=""), dialect="excel")
            headers = ["ts0", "ts1"] + [e.name for e in sc.arr]
            writer.writerow(headers)
            self.csv_writers.append(writer)
            self.arrow_tables.append(sc)
        print("\nInserting data...")

    def finalize(self) -> None:
        print("Done.")

    def on_daq_list(self, daq_list_num: int, timestamp0: int, timestamp1: int, measurements: list) -> None:
        sc = self.arrow_tables[daq_list_num]
        writer = self.csv_writers[daq_list_num]
        data = [timestamp0, timestamp1, *measurements]
        writer.writerow(data)

decoder = Decoder(args.xmraw_file)
decoder.run()



#!/usr/bin/env python
"""Very basic hello-world example.
"""
from pprint import pprint

from pyxcp.cmdline import ArgumentParser
from pyxcp.utils import decode_bytes


daq_info = False


def callout(master, args):
    global daq_info
    if args.daq_info:
        daq_info = True


ap = ArgumentParser(description="pyXCP hello world.", callout=callout)
ap.parser.add_argument(
    "-d",
    "--daq-info",
    dest="daq_info",
    help="Display DAQ-info",
    default=False,
    action="store_true",
)

with ap.run() as x:
    x.connect()
    if x.slaveProperties.optionalCommMode:
        x.getCommModeInfo()
    identifier = x.identifier(0x01)
    print("\nSlave Properties:")
    print("=================")
    print(f"ID: {identifier!r}")
    pprint(x.slaveProperties)
    x.cond_unlock()
    cps = x.getCurrentProtectionStatus()
    print("\nProtection Status")
    print("=================")
    for k, v in cps.items():
        print(f"    {k:6s}: {v}")
    daq = x.getDaqInfo()
    pprint(daq)
    if daq_info:
        dqp = x.getDaqProcessorInfo()
        print("\nDAQ Processor Info:")
        print("===================")
        print(dqp)
        print("\nDAQ Events:")
        print("===========")
        for idx in range(dqp.maxEventChannel):
            evt = x.getDaqEventInfo(idx)
            length = evt.eventChannelNameLength
            name = decode_bytes(x.pull(length))
            dq = "DAQ" if evt.daqEventProperties.daq else ""
            st = "STIM" if evt.daqEventProperties.stim else ""
            dq_st = dq + " " + st
            print(f"    [{idx:04}] {name:r}")
            print(f"        dir:            {dq_st}")
            print(f"        packed:         {evt.daqEventProperties.packed}")
            PFX_CONS = "CONSISTENCY_"
            print(f"        consistency:    {evt.daqEventProperties.consistency.strip(PFX_CONS)}")
            print(f"        max. DAQ lists: {evt.maxDaqList}")
            PFX_TU = "EVENT_CHANNEL_TIME_UNIT_"
            print(f"        unit:           {evt.eventChannelTimeUnit.strip(PFX_TU)}")
            print(f"        cycle:          {evt.eventChannelTimeCycle or 'SPORADIC'}")
            print(f"        priority        {evt.eventChannelPriority}")

        dqr = x.getDaqResolutionInfo()
        print("\nDAQ Resolution Info:")
        print("====================")
        print(dqr)
        for idx in range(dqp.maxDaq):
            print(f"\nDAQ List Info #{idx}")
            print("=================")
            print(f"{x.getDaqListInfo(idx)}")
    x.disconnect()



#!/usr/bin/env python
"""Create / convert pyxcp profiles (configurations).
"""

import sys

from pyxcp.cmdline import ArgumentParser


def main():
    if len(sys.argv) == 1:
        sys.argv.append("profile")
    elif len(sys.argv) >= 2 and sys.argv[1] != "profile":
        sys.argv.insert(1, "profile")

    ap = ArgumentParser(description="Create / convert pyxcp profiles (configurations).")

    try:
        with ap.run() as x:  # noqa: F841
            pass
    except FileNotFoundError as e:
        print(f"Error: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import selectors
import socket
import struct
import threading
from collections import deque
from typing import Optional

from pyxcp.transport.base import BaseTransport
from pyxcp.utils import short_sleep


DEFAULT_XCP_PORT = 5555
RECV_SIZE = 8196


def socket_to_str(sock: socket.socket) -> str:
    peer = sock.getpeername()
    local = sock.getsockname()
    AF = {
        socket.AF_INET: "AF_INET",
        socket.AF_INET6: "AF_INET6",
    }
    TYPE = {
        socket.SOCK_DGRAM: "SOCK_DGRAM",
        socket.SOCK_STREAM: "SOCK_STREAM",
    }
    family = AF.get(sock.family, "OTHER")
    typ = TYPE.get(sock.type, "UNKNOWN")
    res = f"XCPonEth - Connected to: {peer[0]}:{peer[1]}  local address: {local[0]}:{local[1]} [{family}][{typ}]"
    return res


class Eth(BaseTransport):
    """"""

    MAX_DATAGRAM_SIZE = 512
    HEADER = struct.Struct("<HH")
    HEADER_SIZE = HEADER.size

    def __init__(self, config=None, policy=None, transport_layer_interface: Optional[socket.socket] = None) -> None:
        super().__init__(config, policy, transport_layer_interface)
        self.load_config(config)
        self.host: str = self.config.host
        self.port: int = self.config.port
        self.protocol: int = self.config.protocol
        self.ipv6: bool = self.config.ipv6
        self.use_tcp_no_delay: bool = self.config.tcp_nodelay
        address_to_bind: str = self.config.bind_to_address
        bind_to_port: int = self.config.bind_to_port
        self._local_address = (address_to_bind, bind_to_port) if address_to_bind else None
        if self.ipv6 and not socket.has_ipv6:
            msg = "XCPonEth - IPv6 not supported by your platform."
            self.logger.critical(msg)
            raise RuntimeError(msg)
        else:
            address_family = socket.AF_INET6 if self.ipv6 else socket.AF_INET
        proto = socket.SOCK_STREAM if self.protocol == "TCP" else socket.SOCK_DGRAM
        if self.host.lower() == "localhost":
            self.host = "::1" if self.ipv6 else "localhost"

        try:
            addrinfo = socket.getaddrinfo(self.host, self.port, address_family, proto)
            (
                self.address_family,
                self.socktype,
                self.proto,
                self.canonname,
                self.sockaddr,
            ) = addrinfo[0]
        except BaseException as ex:  # noqa: B036
            msg = f"XCPonEth - Failed to resolve address {self.host}:{self.port}"
            self.logger.critical(msg)
            raise Exception(msg) from ex
        self.status: int = 0
        self.sock = socket.socket(self.address_family, self.socktype, self.proto)
        self.selector = selectors.DefaultSelector()
        self.selector.register(self.sock, selectors.EVENT_READ)
        self.use_tcp = self.protocol == "TCP"
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if self.use_tcp and self.use_tcp_no_delay:
            self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        if hasattr(socket, "SO_REUSEPORT"):
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self.sock.settimeout(0.5)
        if self._local_address:
            try:
                self.sock.bind(self._local_address)
            except BaseException as ex:  # noqa: B036
                msg = f"XCPonEth - Failed to bind socket to given address {self._local_address}"
                self.logger.critical(msg)
                raise Exception(msg) from ex
        self._packet_listener = threading.Thread(
            target=self._packet_listen,
            args=(),
            kwargs={},
        )
        self._packets = deque()

    def connect(self) -> None:
        if self.status == 0:
            self.sock.connect(self.sockaddr)
            self.logger.info(socket_to_str(self.sock))
            self.start_listener()
            self.status = 1  # connected

    def start_listener(self) -> None:
        super().start_listener()
        if self._packet_listener.is_alive():
            self._packet_listener.join()
        self._packet_listener = threading.Thread(target=self._packet_listen)
        self._packet_listener.start()

    def close(self) -> None:
        """Close the transport-layer connection and event-loop."""
        self.finish_listener()
        if self.listener.is_alive():
            self.listener.join()
        if self._packet_listener.is_alive():
            self._packet_listener.join()
        self.close_connection()

    def _packet_listen(self) -> None:
        use_tcp: bool = self.use_tcp
        EVENT_READ = selectors.EVENT_READ
        close_event_set = self.closeEvent.is_set
        socket_fileno = self.sock.fileno
        select = self.selector.select
        _packets = self._packets
        if use_tcp:
            sock_recv = self.sock.recv
        else:
            sock_recv = self.sock.recvfrom
        while True:
            try:
                if close_event_set() or socket_fileno() == -1:
                    return
                sel = select(0.02)
                for _, events in sel:
                    if events & EVENT_READ:
                        recv_timestamp = self.timestamp.value
                        if use_tcp:
                            response = sock_recv(RECV_SIZE)
                            if not response:
                                self.sock.close()
                                self.status = 0
                                break
                            else:
                                _packets.append((response, recv_timestamp))
                        else:
                            response, _ = sock_recv(Eth.MAX_DATAGRAM_SIZE)
                            if not response:
                                self.sock.close()
                                self.status = 0
                                break
                            else:
                                _packets.append((response, recv_timestamp))
            except BaseException:  # noqa: B036
                self.status = 0  # disconnected
                break

    def listen(self) -> None:
        HEADER_UNPACK_FROM = self.HEADER.unpack_from
        HEADER_SIZE = self.HEADER_SIZE
        process_response = self.process_response
        popleft = self._packets.popleft
        close_event_set = self.closeEvent.is_set
        socket_fileno = self.sock.fileno
        _packets = self._packets
        length: Optional[int] = None
        counter: int = 0
        data: bytearray = bytearray(b"")
        while True:
            if close_event_set() or socket_fileno() == -1:
                return
            count: int = len(_packets)
            if not count:
                short_sleep()
                continue
            for _ in range(count):
                bts, timestamp = popleft()
                data += bts
                current_size: int = len(data)
                current_position: int = 0
                while True:
                    if length is None:
                        if current_size >= HEADER_SIZE:
                            length, counter = HEADER_UNPACK_FROM(data, current_position)
                            current_position += HEADER_SIZE
                            current_size -= HEADER_SIZE
                        else:
                            data = data[current_position:]
                            break
                    else:
                        if current_size >= length:
                            response = data[current_position : current_position + length]
                            process_response(response, length, counter, timestamp)
                            current_size -= length
                            current_position += length
                            length = None
                        else:
                            data = data[current_position:]
                            break

    def send(self, frame) -> None:
        self.pre_send_timestamp = self.timestamp.value
        self.sock.send(frame)
        self.post_send_timestamp = self.timestamp.value

    def close_connection(self) -> None:
        if not self.invalidSocket:
            # Seems to be problematic /w IPv6
            # if self.status == 1:
            #     self.sock.shutdown(socket.SHUT_RDWR)
            self.sock.close()

    @property
    def invalidSocket(self) -> bool:
        return not hasattr(self, "sock") or self.sock.fileno() == -1
